//
//	Configuration file handling / parsing code
//
//        Copyright (c) 2002-2011 Jim Peters <http://uazu.net/>.
//        Released under the GNU GPL version 2 as published by the
//        Free Software Foundation.  See the file COPYING for details,
//        or visit <http://www.gnu.org/copyleft/gpl.html>.
//

typedef struct Parse Parse;
struct Parse {
   char *txt;		// Data to parse
   char *pos;		// Current read-position
   char *rew;		// 'Rewind' read-position; previous value of ->pos
   			//   after a successful parse()
};

//  //
//  //      Load a file as a single string into memory
//  //
//  
//  static char *
//  load_file(char *fnam) {
//     int len;
//     char *rv;
//     FILE *in;
//  
//     if (!(in= fopen(fnam, "rb")))        // "b" binary for Windows
//        error("Can't open file: %s", fnam);
//  
//     if (0 != fseek(in, 0, SEEK_END) ||
//         0 > (len= ftell(in)))
//        error("Can't find file length: %s", fnam);
//  
//     fseek(in, 0, SEEK_SET);
//  
//     rv= Alloc(len+1);
//     if (1 != fread(rv, len, 1, in) ||
//         0 != fclose(in))
//        error("Read error on file: %s", fnam);
//  
//     rv[len]= 0;
//     return rv;
//  }
//  
//  //
//  //	Report an error on a particular line of the parse text, and
//  //	return 1.
//  //
//  
//  static int 
//  line_error(Parse *pp, char *p, char *fmt, ...) {
//     va_list ap;
//     char buf[1024];
//     int len;
//     va_start(ap, fmt);
//  
//     len= vsnprintf(buf, sizeof(buf), fmt, ap);
//     if (len < 0 || len >= sizeof(buf)-1) 
//        error("line_error exceeded buffer");
//  
//     if (p && p > pp->txt) {
//        int lin= 1;
//        char *q= p, tmp;
//        while (q >= pp->txt) lin += (*q-- == '\n');
//        while (p > pp->txt && p[-1] != '\n') p--;
//        for (q= p; *q && *q != '\n'; q++) ;
//        tmp= *q; *q= 0;
//        applog("Error in config, line %d: %s\n  %s", lin, buf, p);
//        *q= tmp;
//     } else {
//        applog("Error in config: %s", buf);
//     }
//  
//     return 1;
//  }


//
//	Parsing stuff
//

//
//	Attempt to parse from the current position using the format
//	string given.  Tokens in the input are separated by white
//	space, or terminated by a ';' or ','.  Expects any text
//	provided in the formatting string to match tokens exactly.
//	Formatting string white space can match any amount of white
//	space in the input (including newlines characters).  Supported
//	formats:
//
//	%d decimal integer (int*)
//	%f floating-point number (double*)
//	%x hexadecimal integer (int*)
//	%t token string (char**,char**: start pointer, end+1 pointer)
//	%q quoted string (char**,char**: start pointer, end+1 pointer)
//	%r remainder of expression up to ';' (char**,char**: start, end+1)
//	%i identifier [a-zA-Z_][a-zA-Z0-9_]* (char**,char**: start, end+1)
//	%T token string (char**, strdup'd string)
//	%Q quoted string (char**, strdup'd string)
//	%R remainder of expression up to ';' (char**, strdup'd string)
//	%I identifier [a-zA-Z_][a-zA-Z0-9_]* (char**, strdup'd string)
//	%E matches end-of-file
//	%F frequency (Hz) (%f, midi number or note-name) (double*)
//	%S sign '+' or '-', gives 1 or -1 (int*)
//
//	Formats may be used within tokens (e.g. F%d/%d).  %t/%q/%r/%i
//	return pointers into the data being parsed.  %q does not
//	handle any embedded escape sequences nor escaped quote marks.
//	However it does handle both "" and '' strings.
//
//	%T/%Q/%R/%I work the same as the lowercase versions, except
//	that they put a strdup'd string in the given location; if the
//	location already contains a non-0 value, this is free'd first.
//
//	%F handles the following forms: 440.0, C4, C#2, C#2^, C#2.5,
//	Bh3, m60, m60.2, m60^.  Apart from the plain frequencies,
//	these are MIDI note names and MIDI note numbers.  Fractions 
//	give fractions of a semitone, and ^ and v may be used in place
//	of a fraction to indicate half a semitone higher or lower.
//
//	Data is written to the provided locations only if the entire
//	parse-string matches.  This means that it is safe to do many
//	'trial' calls to parse(), knowing that only a successful one
//	will update variables.  Also, the current position in the file
//	being parsed is only updated on a full match.
//
//	Returns: 0 not a complete match (pointer not advanced, no
//	variables updated), 1 complete match (pointer advanced,
//	variables filled in).
//

static int 
parse(Parse *pp, char *fmt0, ...) {
   va_list ap;
   char *p, *q;
   double *dp = NULL;
   int *ip = NULL;
   long lval;
   double dval;
   char **cpp;
   int ch, qu;
   char *fmt;
   int upd;

   va_start(ap, fmt0);
   p= pp->pos;
   while (isspace(*p)) p++;
   pp->rew= pp->pos= p;

   // Makes two passes.  First is to check it all matches, second is
   // to fill in the variables.
   for (upd= 0; upd < 2; upd++) {
      p= pp->pos;
      fmt= fmt0;
      while ((ch= *fmt++)) {
	 if (isspace(ch)) {
	    // This marks a token break, so check that it is valid
	    if (!(*p == ';' || *p == ',' || *p == 0 || isspace(*p)))
	       return 0;
	    while (isspace(*p)) p++;
	    while (isspace(*fmt)) fmt++;
	    continue;
	 }
	 
	 if (ch == ';' || ch == ',') {
	    while (isspace(*p)) p++;
	    while (isspace(*fmt)) fmt++;
	    if (*p++ != ch)
	       return 0;
	    continue;
	 }
	 
	 if (ch == '%') {
	    switch (*fmt++) {
	     case 'd':
		if (upd) ip= va_arg(ap, int*);
		lval= strtol(p, &q, 10);
		if (q == p) return 0;
		if (upd) *ip= lval; 
		p= q; break;
	     case 'f':
		if (upd) dp= va_arg(ap, double*);
		dval= strtod(p, &q);
		if (q == p) return 0;
		if (upd) *dp= dval;
		p= q; break;
	     case 'x':
		if (upd) ip= va_arg(ap, int*);
		lval= strtol(p, &q, 16);
		if (q == p) return 0;
		if (upd) *ip= lval;
		p= q; break;
	     case 't':
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		for (q= p; !(isspace(*p) || *p == ';'  || *p == ',' || *p == 0); p++) ;
		if (q == p) return 0;
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		break;
	     case 'i':
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		if (!(isalpha(*p) || *p == '_')) return 0;
		for (q= p+1; isalnum(*p) || *p == '_'; p++) ;
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		break;
	     case 'q':
		qu= *p++;
		if (qu != '"' && qu != '\'') return 0;
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		while (!(*p == qu || *p == 0)) p++;
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		if (*p++ != qu) return 0;
		break;
	     case 'r':
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		while (!(*p == ';' || *p == ',' || *p == 0)) p++;
		if (upd) { cpp= va_arg(ap, char**); *cpp= p; }
		break;
	     case 'T':
		for (q= p; !(isspace(*p) || *p == ';' || *p == ',' || *p == 0); p++) ;
		if (q == p) return 0;
		if (upd) { cpp= va_arg(ap, char**); free(*cpp); *cpp= StrDupRange(q, p); }
		break;
	     case 'I':
		if (!(isalpha(*p) || *p == '_')) return 0;
		for (q= p+1; isalnum(*p) || *p == '_'; p++) ;
		if (upd) { cpp= va_arg(ap, char**); free(*cpp); *cpp= StrDupRange(q, p); }
		break;
	     case 'Q':
		qu= *p++;
		if (qu != '"' && qu != '\'') return 0;
		for (q= p; !(*p == qu || *p == 0); p++) ;
		if (upd) { cpp= va_arg(ap, char**); free(*cpp); *cpp= StrDupRange(q, p); }
		if (*p++ != qu) return 0;
		break;
	     case 'R':
		for (q= p; !(*p == ';' || *p == ',' || *p == 0); p++) ;
		if (upd) { cpp= va_arg(ap, char**); free(*cpp); *cpp= StrDupRange(q, p); }
		break;
	     case 'E':
		if (*p) return 0;
		break;
	     case 'S':
		if (*p != '+' && *p != '-') return 0;
		p++;
		if (upd) { ip= va_arg(ap, int*); *ip= p[-1] == '+' ? 1 : -1; }
		break;
	     case 'F':
		if (upd) dp= va_arg(ap, double*);
		dval= strtod(p, &q);
		if (q != p) 
		   p= q;
		else { 
		   if (p[0] == 'm') {
		      p++; dval= (double)strtol(p, &q, 10);
		      if (q == p) return 0;
		      p= q;
		   } else if (p[0] >= 'A' && p[0] <= 'G') {
		      dval= "9;02457"[*p++ - 'A'] + 12.0;
		      if (*p == '#') { dval++; p++; }
		      else if (*p == 'b') { dval--; p++; }
		      lval= strtol(p, &q, 10);
		      if (q == p) return 0;
		      p= q; dval += lval * 12.0;
		   } else return 0;
		   if (p[0] == '^') dval += 0.5;
		   else if (p[0] == 'v') dval -= 0.5;
		   else if (p[0] == '.') {
		      dval += strtod(p, &q);
		      if (q == p) return 0;
		      p= q;
		   }
		   dval= 440 * exp(M_LN2 * (dval - 69));    // A4 (m69) is 440Hz
		}
		if (upd) *dp= dval;
		break;
	     default:
		error("Bad format in parse(): unknown sequence %%%c", fmt[-1]);
	    }
	    continue;
	 }
	 
	 if (*p++ != ch)
	    return 0;
      }
   }

   pp->pos= p;
   return 1;
}

//
//	Check that parse is complete, i.e. it has reached the end of
//	the input string.
//

static int 
parseEND(Parse *pp) {
   while (isspace(*pp->pos)) pp->pos++;
   return *pp->pos == 0;
}

//
//	Initialise a parse structure ready to parse the given string
//

static void 
parse_init(Parse *pp, char *txt) {
   pp->txt= txt;
   pp->pos= txt;
   pp->rew= txt;
}

// END //
