#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

// For sleep/Sleep
#ifdef T_WIN32
#include <windows.h>
#include <winbase.h>
#else
#include <unistd.h>
#endif

#include "sbglib.h"

#define RATE 44100

// WARNING!  Cheap code!  For testing only.
// 
// Reads list of specifications from first file on command line and
// feeds them directly to sbg_set().  Outputs using PortAudio by
// default, or to "out.wav" if option -W is used.

// Input format is <chan> <time-ms> [~]<spec>
// ~ for slide-to.
// <spec> is - for 'off', or else binaural spec.
// Channel numbers are arbitrary.
// No blank lines, no comments.

static void 
error(char *fmt, ...) {
   va_list ap; va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");
   exit(1);
}

FILE *out;

void 
writeOut(char *buf, int siz) {
  if (1 != fwrite(buf, siz, 1, out))
     error("Write error");
}         

//
//	Write a WAV header, and setup out_mode if byte-swapping is
//	required.  'byte_count' should have been set up by this point.
//

#define addU4(xx) { int a= xx; *p++= a; *p++= (a >>= 8); *p++= (a >>= 8); *p++= (a >>= 8); }
#define addStr(xx) { char *q= xx; *p++= *q++; *p++= *q++; *p++= *q++; *p++= *q++; }

void 
writeWavHeader(int cnt) {
  char buf[44], *p= buf;
  int byte_count = cnt * 4;
  int out_rate = RATE;
  int out_bps = 4;

  addStr("RIFF");
  addU4(byte_count + 36);
  addStr("WAVE");
  addStr("fmt ");
  addU4(16);
  addU4(0x00020001);
  addU4(out_rate);
  addU4(out_rate * out_bps);
  addU4(0x0004 + 0x10000*(out_bps*4));	// 2,4 -> 8,16 - always assume stereo
  addStr("data");
  addU4(byte_count);

  writeOut(buf, 44);
}


int main(int ac, char **av) {
   SBG *sbg = sbg_init(RATE);
   int chan;
   int time;
   char spec[1024];
   char line[1024];
   int maxtime = 0;
   FILE *in;
   int wav_out = 0;
   int debug_dump = 0;
   char *err;

   ac--, av++;
   while (ac > 0) {
      if (0 == strcmp(av[0], "-W")) {
         wav_out = 1;
         ac--, av++;
         continue;
      }
      if (0 == strcmp(av[0], "-D")) {
         debug_dump = 1;
         ac--, av++;
         continue;
      }
      break;
   }
   
   if (ac != 1) error("Usage: pa-play [-W] [-D] <input-file>\n"
                      "  -W output out.wav instead of playing\n"
                      "  -D detailed debug dump of events loaded");
   
   in = fopen(av[0], "r");
   if (!in) error("Can't open file: %s", av[0]);

   while (fgets(line, sizeof(line), in)) {
      if (line[0] == '#') {
         printf("%s", line);
         continue;
      }
      if (line[0] == '\n') continue;
      if (3 == sscanf(line, "%d %d %1023s\n", &chan, &time, spec)) {
         int slideto = (spec[0] == '~');
         char *sp = &spec[slideto];
         time *= 44;
         printf("%4d %8d %d %s\n", chan, time/44, slideto, sp);
         err = sbg_set(sbg, chan, time, slideto, sp);
         if (err) error("Invalid line: %s", err);
         if (time > maxtime) maxtime = time;
         continue;
      }
      error("Unexpected line: %s", line);
   }
   fclose(in);

   if (debug_dump)
      sbg_debug_dump(sbg, stdout);

   maxtime += 200 * 44;  // Add 200ms on end

   if (wav_out) {
      out = fopen("out.wav", "wb");
      if (!out) error("Can't create out.wav");
      
      writeWavHeader(maxtime);
      
      while (maxtime > 0) {
         short buf[16384];
         int cnt = sizeof(buf) / sizeof(buf[0]) / 2;
         if (cnt > maxtime) cnt = maxtime;
         sbg_gen(sbg, &buf[0], cnt);
         writeOut((char*)buf, sizeof(buf[0]) * 2 * cnt);
         maxtime -= cnt;
      }
      
      if (0 != fclose(out))
         error("Write failed in close");
   } else {
      char *err;

      err = sbg_pa_start(sbg);
      if (err) error("%s", err);
      
      while (sbg_now(sbg) < maxtime) {
#ifdef T_WIN32
         Sleep(100);
#else
         sleep(1);
#endif
      }
      
      err = sbg_pa_stop();
      if (err) error("%s", err);
   }
      
   sbg_term(sbg);
   
   return 0;
}
   
